@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 2002-2016. All Rights Reserved.
+%% Copyright Ericsson AB 2002-2018. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -23,7 +23,7 @@
 -export([vsn/0]).
 
 %% observer stuff
--export([sys_info/0, get_port_list/0,
+-export([sys_info/0, get_port_list/0, procs_info/1,
 	 get_table/3, get_table_list/2, fetch_stats/2]).
 
 %% etop stuff
@@ -36,6 +36,7 @@
 	 ttb_write_binary/2,
 	 ttb_stop/1,
 	 ttb_fetch/2,
+	 ttb_fetch/3,
          ttb_resume_trace/0,
 	 ttb_get_filenames/1]).
 -define(CHUNKSIZE,8191). % 8 kbytes - 1 byte
@@ -63,9 +64,7 @@ sys_info() ->
 			  end,
 
     {{_,Input},{_,Output}} = erlang:statistics(io),
-    [{process_count, erlang:system_info(process_count)},
-     {process_limit, erlang:system_info(process_limit)},
-     {uptime, element(1, erlang:statistics(wall_clock))},
+    [{uptime, element(1, erlang:statistics(wall_clock))},
      {run_queue, erlang:statistics(run_queue)},
      {io_input, Input},
      {io_output,  Output},
@@ -86,7 +85,17 @@ sys_info() ->
      {thread_pool_size, erlang:system_info(thread_pool_size)},
      {wordsize_internal, erlang:system_info({wordsize, internal})},
      {wordsize_external, erlang:system_info({wordsize, external})},
-     {alloc_info, alloc_info()}
+     {alloc_info, alloc_info()},
+     {process_count, erlang:system_info(process_count)},
+     {atom_limit,  erlang:system_info(atom_limit)},
+     {atom_count, erlang:system_info(atom_count)},
+     {process_limit, erlang:system_info(process_limit)},
+     {process_count, erlang:system_info(process_count)},
+     {port_limit, erlang:system_info(port_limit)},
+     {port_count, erlang:system_info(port_count)},
+     {ets_limit,  erlang:system_info(ets_limit)},
+     {ets_count, erlang:system_info(ets_count)},
+     {dist_buf_busy_limit, erlang:system_info(dist_buf_busy_limit)}
      | MemInfo].
 
 alloc_info() ->
@@ -179,8 +188,8 @@ inet_port_extra({_,Type},Port) when Type =:= "udp_inet";
             {error, _} -> []
         end ++
         case inet:getopts(Port,
-                          [active, broadcast, buffer, delay_send,
-                           deliver, dontroute, exit_on_close,
+                          [active, broadcast, buffer, bind_to_device,
+                           delay_send, deliver, dontroute, exit_on_close,
                            header, high_msgq_watermark, high_watermark,
                            ipv6_v6only, keepalive, linger, low_msgq_watermark,
                            low_watermark, mode, netns, nodelay, packet,
@@ -270,7 +279,7 @@ get_table_list(mnesia, Opts) ->
 			     end,
 		       [Tab|Acc]
 		   catch _:_What ->
-			   %% io:format("Skipped ~p: ~p ~p ~n",[Id, _What, erlang:get_stacktrace()]),
+			   %% io:format("Skipped ~p: ~p ~p ~n",[Id, _What, Stacktrace]),
 			   Acc
 		   end
 	   end,
@@ -284,7 +293,7 @@ fetch_stats_loop(Parent, Time) ->
     erlang:system_flag(scheduler_wall_time, true),
     receive
 	_Msg ->
-	    %% erlang:system_flag(scheduler_wall_time, false)
+	    erlang:system_flag(scheduler_wall_time, false),
 	    ok
     after Time ->
 	    _M = Parent ! {stats, 1,
@@ -293,6 +302,23 @@ fetch_stats_loop(Parent, Time) ->
 			   try erlang:memory() catch _:_ -> [] end},
 	    fetch_stats_loop(Parent, Time)
     end.
+
+%%
+%% Chunk sending process info to etop/observer
+%%
+procs_info(Collector) ->
+    All = processes(),
+    Send = fun Send (Pids) ->
+                   try lists:split(10000, Pids) of
+                       {First, Rest} ->
+                           Collector ! {procs_info, self(), etop_collect(First, [])},
+                           Send(Rest)
+                   catch _:_ ->
+                           Collector ! {procs_info, self(), etop_collect(Pids, [])}
+                   end
+           end,
+    Send(All).
+
 %%
 %% etop backend
 %%
@@ -314,19 +340,18 @@ etop_collect(Collector) ->
 
     case SchedulerWallTime of
 	undefined ->
-	    spawn(fun() -> flag_holder_proc(Collector) end),
+            spawn(fun() -> flag_holder_proc(Collector) end),
             ok;
 	_ ->
 	    ok
-    end,
-
-    erlang:system_flag(scheduler_wall_time,true).
+    end.
 
 flag_holder_proc(Collector) ->
+    erlang:system_flag(scheduler_wall_time,true),
     Ref = erlang:monitor(process,Collector),
     receive
 	{'DOWN',Ref,_,_,_} ->
-	    %% erlang:system_flag(scheduler_wall_time,false)
+	    erlang:system_flag(scheduler_wall_time,false),
 	    ok
     end.
 
@@ -408,7 +433,9 @@ ttb_meta_tracer(MetaFile,PI,Parent,SessionData) ->
 	    ReturnMS = [{'_',[],[{return_trace}]}],
 	    erlang:trace_pattern({erlang,spawn,3},ReturnMS,[meta]),
 	    erlang:trace_pattern({erlang,spawn_link,3},ReturnMS,[meta]),
-	    erlang:trace_pattern({erlang,spawn_opt,1},ReturnMS,[meta]),
+	    erlang:trace_pattern({erlang,spawn_opt,4},ReturnMS,[meta]),
+	    erlang:trace_pattern({erts_internal,spawn_init,1},[],[meta]),
+	    erlang:trace_pattern({erts_internal,dist_spawn_init,1},[],[meta]),
 	    erlang:trace_pattern({erlang,register,2},[],[meta]),
 	    erlang:trace_pattern({global,register_name,2},[],[meta]),
             ok;
@@ -434,7 +461,7 @@ ttb_meta_tracer_loop(MetaFile,PI,Acc,State) ->
 	{trace_ts,_,call,{global,register_name,[Name,Pid]},_} ->
 	    ok = ttb_store_meta({pid,{Pid,{global,Name}}},MetaFile),
 	    ttb_meta_tracer_loop(MetaFile,PI,Acc,State);
-	{trace_ts,CallingPid,call,{erlang,spawn_opt,[{M,F,Args,_}]},_} ->
+	{trace_ts,CallingPid,call,{erlang,spawn_opt,[M,F,Args,_]},_} ->
 	    MFA = {M,F,length(Args)},
 	    NewAcc = dict:update(CallingPid,
 				 fun(Old) -> [MFA|Old] end, [MFA], 
@@ -472,6 +499,16 @@ ttb_meta_tracer_loop(MetaFile,PI,Acc,State) ->
 			    Acc),
 	    ttb_meta_tracer_loop(MetaFile,PI,NewAcc,State);
 
+	{trace_ts,CallingPid,call,{erts_internal,spawn_init,[{M,F,Args}]},_} ->
+            %% Local spawn_request()...
+            ok = ttb_store_meta({pid,{CallingPid,{M,F,length(Args)}}},MetaFile),
+	    ttb_meta_tracer_loop(MetaFile,PI,Acc,State);
+
+	{trace_ts,CallingPid,call,{erts_internal, dist_spawn_init, [MFnoA]},_} ->
+            %% Distributed spawn_request()...
+            ok = ttb_store_meta({pid,{CallingPid,MFnoA}},MetaFile),
+	    ttb_meta_tracer_loop(MetaFile,PI,Acc,State);
+
 	{metadata,Data} when is_list(Data) ->
 	    ok = ttb_store_meta(Data,MetaFile),
 	    ttb_meta_tracer_loop(MetaFile,PI,Acc,State);
@@ -505,7 +542,9 @@ ttb_meta_tracer_loop(MetaFile,PI,Acc,State) ->
             try_stop_overload_check(State),
             erlang:trace_pattern({erlang,spawn,3},false,[meta]),
 	    erlang:trace_pattern({erlang,spawn_link,3},false,[meta]),
-	    erlang:trace_pattern({erlang,spawn_opt,1},false,[meta]),
+	    erlang:trace_pattern({erlang,spawn_opt,4},false,[meta]),
+	    erlang:trace_pattern({erts_internal,spawn_init,1},false,[meta]),
+	    erlang:trace_pattern({erts_internal,dist_spawn_init,1},false,[meta]),
 	    erlang:trace_pattern({erlang,register,2},false,[meta]),
 	    erlang:trace_pattern({global,register_name,2},false,[meta]);
 	stop ->
@@ -634,22 +673,42 @@ stop_seq_trace() ->
 
 %% Fetch ttb logs from remote node
 ttb_fetch(MetaFile,{Port,Host}) ->
+    ttb_fetch(MetaFile,{Port,Host},undefined).
+ttb_fetch(MetaFile,{Port,Host},MasterEnc) ->
     erlang:process_flag(priority,low),
     Files = ttb_get_filenames(MetaFile),
     {ok, Sock} = gen_tcp:connect(Host, Port, [binary, {packet, 2}]),
-    send_files({Sock,Host},Files),
+    send_files({Sock,Host},Files,MasterEnc,file:native_name_encoding()),
     ok = gen_tcp:close(Sock).
 
 
-send_files({Sock,Host},[File|Files]) ->
+send_files({Sock,Host},[File|Files],MasterEnc,MyEnc) ->
     {ok,Fd} = file:open(File,[raw,read,binary]),
-    ok = gen_tcp:send(Sock,<<1,(list_to_binary(filename:basename(File)))/binary>>),
+    Basename = filename:basename(File),
+    {Code,FilenameBin} = encode_filename(Basename,MasterEnc,MyEnc),
+    ok = gen_tcp:send(Sock,<<Code,FilenameBin/binary>>),
     send_chunks(Sock,Fd),
     ok = file:delete(File),
-    send_files({Sock,Host},Files);
-send_files({_Sock,_Host},[]) ->
+    send_files({Sock,Host},Files,MasterEnc,MyEnc);
+send_files({_Sock,_Host},[],_MasterEnc,_MyEnc) ->
     done.
 
+encode_filename(Basename,undefined,MyEnc) ->
+    %% Compatible with old version of ttb.erl, but no longer crashing
+    %% for code points > 255.
+    {1,unicode:characters_to_binary(Basename,MyEnc,MyEnc)};
+encode_filename(Basename,MasterEnc,MyEnc) ->
+    case unicode:characters_to_binary(Basename,MyEnc,MasterEnc) of
+        Bin when is_binary(Bin) ->
+            %% Encoding succeeded
+            {2,Bin};
+        _ ->
+            %% Can't convert Basename from my encoding to the master
+            %% node's encoding. Doing my best and hoping that master
+            %% node can fix it...
+            {3,unicode:characters_to_binary(Basename,MyEnc,MyEnc)}
+    end.
+
 send_chunks(Sock,Fd) ->
     case file:read(Fd,?CHUNKSIZE) of
 	{ok,Bin} -> 
@@ -707,6 +766,7 @@ sys_tables() ->
      mnesia_gvar, mnesia_stats,
 						%     mnesia_transient_decision,
      pg2_table,
+     pg,
      queue,
      schema,
      shell_records,
@@ -718,7 +778,7 @@ sys_tables() ->
 
 sys_processes() ->
     [auth, code_server, global_name_server, inet_db,
-     mnesia_recover, net_kernel, timer_server, wxe_master].
+     mnesia_recover, net_kernel, pg, timer_server, wxe_master].
 
 mnesia_tables() ->
     [ir_AliasDef, ir_ArrayDef, ir_AttributeDef, ir_ConstantDef,
